//
// Copyright (c) 2022, MindFusion LLC - Bulgaria.
//

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;

using MindFusion.Diagramming.Layout;
using MindFusion.Scripting;
using System.Xml;


namespace MindFusion.Diagramming.WinForms.Samples.CS.Scripting
{
	public partial class MainForm : Form
	{
		static MainForm()
		{
			Diagram.RegisterItemClass(typeof(VariableListNode), "cnVarList", 1);
			Diagram.RegisterItemClass(typeof(OperationNode), "cnOperation", 1);
			Diagram.RegisterItemClass(typeof(TerminatorNode), "cnTerminator", 1);
			Diagram.RegisterItemClass(typeof(FunctionNode), "cnFunction", 1);
		}

		public MainForm()
		{
			InitializeComponent();

			diagramView.Behavior = Behavior.DrawLinks;
			diagramView.ModificationStart = ModificationStart.AutoHandles;
			diagramView.MouseDown += OnDiagramViewMouseDown;
			diagramView.ScrollChanged += OnDiagramViewScrollChanged;
			diagramView.SizeChanged += OnDiagramViewSizeChanged;

			diagram.RouteLinks = true;
			//diagram.RoutingOptions.CrossingCost = 255;
			//diagram.LinkRouter = new GridRouter();
			diagram.ShowAnchors = ShowAnchors.Always;
			diagram.RoundedLinks = true;
			diagram.SelectionOnTop = false;
			diagram.LinkShape = LinkShape.Cascading;
			diagram.LinkHeadShape = ArrowHeads.Triangle;
			diagram.LinkHeadShapeSize = 3;
			diagram.DiagramLinkStyle.Brush = new MindFusion.Drawing.SolidBrush(Color.Black);

			diagram.NodeSelected += OnNodeSelected;
			diagram.NodeDeselected += OnNodeDeselected;
			diagram.DrawAnchorPoint += OnDrawAnchorPoint;
			diagram.DrawLink += OnDrawLink;
			diagram.LinkCreating += OnLinkCreating;
			diagram.NodeCreated += OnNodeCreated;

			diagram.ShadowOffsetX = 0;
			diagram.ShadowOffsetY = 0;
			diagram.LinkCustomDraw = CustomDraw.ShadowOnly;
			diagram.LinkHandlesStyle = HandlesStyle.Custom;

			// Create the special variable list node
			variableListNode = new VariableListNode(diagram);
			variableListNode.Visible = false;
			variableListNode.Resized += OnVariableListNodeResized;
			diagram.Nodes.Add(variableListNode);
			UpdateVariableListPosition();

			evaluationContext = new EvaluationContext(this);
			evaluationContext.DispatchFunctionCall += OnDispatchFunctionCall;

			examplesListBox.Items.Add("Factorial");
			examplesListBox.Items.Add("Gcd");
			examplesListBox.Items.Add("Fibonacci");
		}

		private void OnDiagramViewScrollChanged(object sender, EventArgs e)
		{
			UpdateVariableListPosition();
		}

		private void OnDiagramViewSizeChanged(object sender, EventArgs e)
		{
			UpdateVariableListPosition();
		}

		private void OnVariableListNodeResized(object sender, EventArgs e)
		{
			UpdateVariableListPosition();
		}

		private void UpdateVariableListPosition()
		{
			Rectangle clientViewport = diagramView.ClientRectangle;
			clientViewport.Width -= SystemInformation.VerticalScrollBarWidth;
			clientViewport.Height -= SystemInformation.HorizontalScrollBarHeight;
			RectangleF viewport = diagramView.ClientToDoc(clientViewport);

			variableListNode.Bounds = new RectangleF(
				viewport.Right - variableListNode.Bounds.Width - 2,
				viewport.Bottom  - variableListNode.Bounds.Height - 2,
				variableListNode.Bounds.Width, variableListNode.Bounds.Height);
		}

		private void OnDiagramViewMouseDown(object sender, MouseEventArgs e)
		{
			diagram.FocusedComponent = null;
		}

		private void OnNodeSelected(object sender, NodeEventArgs e)
		{
			ScriptedNode node = e.Node as ScriptedNode;
			if (node != null)
				node.OnSelected();
		}

		private void OnNodeDeselected(object sender, NodeEventArgs e)
		{
			ScriptedNode node = e.Node as ScriptedNode;
			if (node != null)
				node.OnDeselected();
		}

		private void OnDrawAnchorPoint(object sender, DrawAnchorPointEventArgs e)
		{
			const float size = 1.5f;
			RectangleF bounds = new RectangleF(e.Location.X - size, e.Location.Y - size, 2 * size, 2 * size);
			AnchorPoint point = e.AnchorPattern.Points[e.AnchorIndex];

			switch ((int)point.Tag)
			{

				case 0: // Top-in
				case 3: // Bottom-out
					{
						PointF[] triangle = new PointF[]
						{
							new PointF(bounds.Left, bounds.Top),
							new PointF(bounds.Left + bounds.Width / 2, bounds.Bottom),
							new PointF(bounds.Right, bounds.Top),
							new PointF(bounds.Left, bounds.Top),
						};

						using (Brush brush = new SolidBrush(Color.FromArgb(100, Color.Black)))
							e.Graphics.FillPolygon(brush, triangle);
						using (Pen pen = new Pen(Color.Black, 0))
							e.Graphics.DrawPolygon(pen, triangle);
					}
					break;

				case 1: // True
					{
						using (Brush brush = new SolidBrush(Color.FromArgb(150, Color.Green)))
							e.Graphics.FillEllipse(brush, bounds);
						using (Pen pen = new Pen(Color.Black, 0))
							e.Graphics.DrawEllipse(pen, bounds);

						using (Pen pen = new Pen(Color.White, 0))
						{
							e.Graphics.DrawLine(pen, bounds.Left + 0.75f, bounds.Top + bounds.Height / 2,
								bounds.Right - 0.75f, bounds.Top + bounds.Height / 2);
							e.Graphics.DrawLine(pen, bounds.Left + bounds.Width / 2, bounds.Top + 0.75f,
								bounds.Left + bounds.Width / 2, bounds.Bottom - 0.75f);
						}
					}
					break;

				case 2: // False
					{
						using (Brush brush = new SolidBrush(Color.FromArgb(150, Color.Red)))
							e.Graphics.FillEllipse(brush, bounds);
						using (Pen pen = new Pen(Color.Black, 0))
							e.Graphics.DrawEllipse(pen, bounds);

						using (Pen pen = new Pen(Color.White, 0))
						{
							e.Graphics.DrawLine(pen, bounds.Left + 0.75f, bounds.Top + bounds.Height / 2,
								bounds.Right - 0.75f, bounds.Top + bounds.Height / 2);
						}
					}
					break;

			}
		}

		private void OnDrawLink(object sender, DrawLinkEventArgs e)
		{
			if (e.Link.Selected)
			{
				using (Pen pen = new Pen(Color.LightSteelBlue, 1))
					e.Graphics.DrawLines(pen, e.Points.GetArray());
			}
		}

		private void OnLinkCreating(object sender, LinkValidationEventArgs e)
		{
			// Prevent self loops
			if (e.Origin == e.Destination)
			{
				e.Cancel = true;
				return;
			}

			// Prevent two links going out of the same anchor point
			foreach (DiagramLink link in e.Link.Origin.OutgoingLinks)
			{
				if (link.OriginAnchor == e.Link.OriginAnchor)
				{
					e.Cancel = true;
					return;
				}
			}
		}

		private void OnNodeCreated(object sender, NodeEventArgs e)
		{
			variableListNode.ZTop(false);
		}

		private void OnDispatchFunctionCall(object sender, EvaluateFunctionEventArgs e)
		{
			switch (e.Name)
			{

				case "MessageBox":
					{
						e.Result = null;
						if (e.Arguments.Length > 0 && e.Arguments[0] != null)
							MessageBox.Show(e.Arguments[0].ToString());
					}
					break;

				case "Prompt":
					{
						PromptDialog dialog = new PromptDialog();
						if (e.Arguments.Length > 0 && e.Arguments[0] != null)
							dialog.Text = e.Arguments[0].ToString();
						if (dialog.ShowDialog(this) == DialogResult.OK)
							e.Result = dialog.Value;
						else
							e.Result = null;
					}
					break;

				default:
					{
						// Check if it is a function defined in the diagram
						FunctionNode function = FindFunctionNode(e.Name);
						if (function != null)
						{
							EvaluationContext functionContext = new EvaluationContext(this);
							functionContext.DispatchFunctionCall += OnDispatchFunctionCall;
							string[] parameterNames = function.ParameterNames;

							if (parameterNames == null && e.Arguments.Length != 0 ||
								parameterNames != null && parameterNames.Length != e.Arguments.Length)
								throw new Exception(string.Format("The number of formal and actual arguments of the function {0} do not match.", e.Name));

							if (parameterNames != null)
							{
								for (int i = 0; i < parameterNames.Length; i++)
									functionContext.LocalVars.Add(parameterNames[i], e.Arguments[i]);
							}

							e.Result = Execute(function.Code, functionContext);
						}
					}
					break;

			}
		}

		private void startButton_Click(object sender, EventArgs e)
		{
			diagram.Selection.Clear();

			foreach (DiagramNode n in diagram.Nodes)
			{
				TerminatorNode terminator = n as TerminatorNode;
				if (terminator == null)
					continue;
				if (terminator.Type != TerminatorNodeType.Start)
					continue;

				terminator.Selected = true;
				diagramView.BringIntoView(terminator);

				return;
			}

			TerminatorNode node = new TerminatorNode(diagram);
			node.Bounds = new RectangleF(20 + diagramView.ScrollX, 20 + diagramView.ScrollY, 24, 8);
			node.Type = TerminatorNodeType.Start;
           	node.Selected = true;
			diagram.Nodes.Add(node);
		}

		private void endButton_Click(object sender, EventArgs e)
		{
			diagram.Selection.Clear();

			foreach (DiagramNode n in diagram.Nodes)
			{
				TerminatorNode terminator = n as TerminatorNode;
				if (terminator == null)
					continue;
				if (terminator.Type != TerminatorNodeType.End)
					continue;

				terminator.Selected = true;
				diagramView.BringIntoView(terminator);

				return;
			}

			TerminatorNode node = new TerminatorNode(diagram);
			node.Bounds = new RectangleF(20 + diagramView.ScrollX, 20 + diagramView.ScrollY, 24, 8);
			node.Type = TerminatorNodeType.End;
			node.Selected = true;
			diagram.Nodes.Add(node);
		}

		private void operationButton_Click(object sender, EventArgs e)
		{
			diagram.Selection.Clear();

			OperationNode node = new OperationNode(diagram);
			node.Type = OperationType.Operation;
			node.Bounds = new RectangleF(20 + diagramView.ScrollX, 20 + diagramView.ScrollY, 40, 30);
			node.Selected = true;
			diagram.Nodes.Add(node);
		}

		private void choiceButton_Click(object sender, EventArgs e)
		{
			diagram.Selection.Clear();

			OperationNode node = new OperationNode(diagram);
			node.Type = OperationType.Choice;
			node.Bounds = new RectangleF(20 + diagramView.ScrollX, 20 + diagramView.ScrollY, 40, 30);
			node.Selected = true;
			diagram.Nodes.Add(node);
		}

		private void functionButton_Click(object sender, EventArgs e)
		{
			diagram.Selection.Clear();

			FunctionNode node = new FunctionNode(diagram);
			node.Bounds = new RectangleF(20 + diagramView.ScrollX, 20 + diagramView.ScrollY, 40, 30);
			node.Selected = true;
			diagram.Nodes.Add(node);
		}

		private void clearButton_Click(object sender, EventArgs e)
		{
			for (int i = 0; i < diagram.Nodes.Count; )
			{
				DiagramNode node = diagram.Nodes[i];
				if (node is VariableListNode)
				{
					i++;
					continue;
				}

				diagram.Nodes.Remove(diagram.Nodes[i]);
			}
		}

		private void validateButton_Click(object sender, EventArgs e)
		{
			if (ValidateProcess())
			{
				MessageBox.Show(this, "Validation successful!", "Validation",
					MessageBoxButtons.OK, MessageBoxIcon.Information);
			}
		}

		private void runButton_Click(object sender, EventArgs e)
		{
			variableListNode.Clear();

			if (!ValidateProcess())
				return;

			ScriptedNode node = FindStartNode();

			while (true)
			{
				try
				{
					node = Process(node);
				}
				catch (Exception ex)
				{
					MessageBox.Show(this, ex.Message, "Execution Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
					return;
				}

				if (node == null)
					break;
			}
		}

		private void stepRunButton_Click(object sender, EventArgs e)
		{
			if (!tracing)
			{
				variableListNode.Clear();

				if (!ValidateProcess())
					return;

				tracing = true;

				// Update UI
				startButton.Enabled = false;
				operationButton.Enabled = false;
				choiceButton.Enabled = false;
				endButton.Enabled = false;
				functionButton.Enabled = false;
				clearButton.Enabled = false;
				validateButton.Enabled = false;
				runButton.Enabled = false;
				saveButton.Enabled = false;
				loadButton.Enabled = false;

				stepRunButton.Text = "Next";
				stopButton.Enabled = true;

				variableListNode.Visible = true;

				currentNode = FindStartNode();
				currentNode.ShowHighlight();
				diagramView.BringIntoView(currentNode);
			}
			else
			{
				try
				{
					ScriptedNode previousNode = currentNode;
					currentNode = Process(previousNode);

					previousNode.HideHighlight();
					if (currentNode != null)
					{
						currentNode.ShowHighlight();
						diagramView.BringIntoView(currentNode);
					}
					else
					{
						StopRun();
					}
				}
				catch (Exception ex)
				{
					MessageBox.Show(this, ex.Message, "Execution Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
					StopRun();
				}
			}
		}

		private void stopButton_Click(object sender, EventArgs e)
		{
			StopRun();
		}

		private void StopRun()
		{
			variableListNode.Visible = false;

			ClearHighlight();

			startButton.Enabled = true;
			operationButton.Enabled = true;
			choiceButton.Enabled = true;
			endButton.Enabled = true;
			functionButton.Enabled = true;
			clearButton.Enabled = true;
			validateButton.Enabled = true;
			runButton.Enabled = true;
			saveButton.Enabled = true;
			loadButton.Enabled = true;

			stepRunButton.Text = "Step by step";
			stopButton.Enabled = false;

			tracing = false;
		}

		private void saveButton_Click(object sender, EventArgs e)
		{
			SaveFileDialog dialog = new SaveFileDialog();
			dialog.Filter = "XML Files (*.xml)|*.xml|All Files (*.*)|*.*||";

			if (dialog.ShowDialog(this) == DialogResult.OK)
				diagram.SaveToXml(dialog.FileName);
		}

		private void loadButton_Click(object sender, EventArgs e)
		{
			OpenFileDialog dialog = new OpenFileDialog();
			dialog.Filter = "XML Files (*.xml)|*.xml|All Files (*.*)|*.*||";

			if (dialog.ShowDialog(this) == DialogResult.OK)
			{
				diagram.LoadFromXml(dialog.FileName);

				OnDiagramLoaded();
			}
		}

		private void OnDiagramLoaded()
		{
			// Obtain a reference to the new variable list node
			foreach (DiagramNode node in diagram.Nodes)
			{
				VariableListNode variableList = node as VariableListNode;
				if (variableList != null)
				{
					variableListNode = variableList;
					variableListNode.Resized += OnVariableListNodeResized;
					variableListNode.ZTop(false);
					break;
				}
			}
		}

		private bool ValidateProcess()
		{
			// Check for connectivity
			MindFusion.Graphs.Graph graph = new MindFusion.Graphs.Graph();
			Dictionary<DiagramNode, MindFusion.Graphs.Vertex> vertexMap =
				new Dictionary<DiagramNode, MindFusion.Graphs.Vertex>();
			Dictionary<DiagramLink, MindFusion.Graphs.Edge> edgeMap =
				new Dictionary<DiagramLink, MindFusion.Graphs.Edge>();
			foreach (DiagramNode node in diagram.Nodes)
			{
				if (node == variableListNode)
					continue;

				if (node is FunctionNode)
					continue;

				vertexMap.Add(node, graph.AddVertex());
			}

			foreach (DiagramLink link in diagram.Links)
			{
				MindFusion.Graphs.Vertex origin = vertexMap[link.Origin];
				MindFusion.Graphs.Vertex destination = vertexMap[link.Destination];

				edgeMap.Add(link, graph.AddEdge(origin, destination));
			}

			if (graph.GetConnectedComponents().Length > 1)
			{
				MessageBox.Show(this, "Validation failed! The process contains disconnected components.",
					"Validation", MessageBoxButtons.OK, MessageBoxIcon.Error);
				return false;
			}

			// Check for start/end nodes
			bool hasStart = false;
			bool hasEnd = false;
			foreach (DiagramNode node in diagram.Nodes)
			{
				TerminatorNode terminator = node as TerminatorNode;
				if (terminator == null)
					continue;

				if (terminator.Type == TerminatorNodeType.Start)
					hasStart = true;
				else if (terminator.Type == TerminatorNodeType.End)
					hasEnd = true;
			}

			if (!hasStart)
			{
				MessageBox.Show(this, "Validation failed! The process does not have a start node.",
					"Validation", MessageBoxButtons.OK, MessageBoxIcon.Error);
				return false;
			}

			if (!hasEnd)
			{
				MessageBox.Show(this, "Validation failed! The process does not have an end node.",
					"Validation", MessageBoxButtons.OK, MessageBoxIcon.Error);
				return false;
			}

			// Check for missing links - all out anchors must have links
			foreach (DiagramNode node in diagram.Nodes)
			{
				OperationNode operation = node as OperationNode;
				if (operation == null)
					continue;

				if (operation.Type == OperationType.Choice)
				{
					if (operation.IncomingLinks.Count == 0)
					{
						diagram.Selection.Change(operation);
						MessageBox.Show(this, "Validation failed! A choice has an undefined input.",
							"Validation", MessageBoxButtons.OK, MessageBoxIcon.Error);
						return false;
					}
					if (operation.OutgoingLinks.Count != 2)
					{
						diagram.Selection.Change(operation);
						MessageBox.Show(this, "Validation failed! A choice has an undefined output.",
							"Validation", MessageBoxButtons.OK, MessageBoxIcon.Error);
						return false;
					}
				}
				else if (operation.Type == OperationType.Operation)
				{
					if (operation.IncomingLinks.Count == 0)
					{
						diagram.Selection.Change(operation);
						MessageBox.Show(this, "Validation failed! An operation has an undefined input.",
							"Validation", MessageBoxButtons.OK, MessageBoxIcon.Error);
						return false;
					}
					if (operation.OutgoingLinks.Count != 1)
					{
						diagram.Selection.Change(operation);
						MessageBox.Show(this, "Validation failed! An operation has an undefined output.",
							"Validation", MessageBoxButtons.OK, MessageBoxIcon.Error);
						return false;
					}
				}
			}

			return true;
		}

		private ScriptedNode Process(ScriptedNode node)
		{
			TerminatorNode terminator = node as TerminatorNode;
			if (terminator != null)
			{
				if (terminator.Type == TerminatorNodeType.End)
					return null;

				if (terminator.Type == TerminatorNodeType.Start)
					return terminator.OutgoingLinks[0].Destination as ScriptedNode;

				throw new Exception("Invalid terminator.");
			}

			OperationNode operation = node as OperationNode;
			if (operation != null)
			{
				string code = operation.Code;
				object result = Execute(code);

				if (operation.Type == OperationType.Operation)
					return operation.OutgoingLinks[0].Destination as ScriptedNode;

				if (operation.Type == OperationType.Choice)
				{
					if (result == null)
						throw new Exception("Failed to evaluate choice.");

					if (System.Convert.ToBoolean(result))
						return FindTrueLink(operation).Destination as ScriptedNode;
					else
						return FindFalseLink(operation).Destination as ScriptedNode;
				}
			}

			throw new Exception("Invalid process.");
		}

		private object Execute(string expression)
		{
			return Execute(expression, evaluationContext);
		}

		private object Execute(string expression, EvaluationContext context)
		{
			ByteCode code = Compiler.Compile(expression, true);
			object result = code.Evaluate(context);

			// Update variables
			variableListNode.Clear();
			foreach (string varName in context.LocalVars.Keys)
				variableListNode.RegisterVariable(varName, context.LocalVars[varName]);

			return result;
		}

		private TerminatorNode FindStartNode()
		{
			foreach (DiagramNode node in diagram.Nodes)
			{
				TerminatorNode terminator = node as TerminatorNode;
				if (terminator != null)
				{
					if (terminator.Type == TerminatorNodeType.Start)
						return terminator;
				}
			}

			return null;
		}

		private FunctionNode FindFunctionNode(string name)
		{
			foreach (DiagramNode node in diagram.Nodes)
			{
				FunctionNode function = node as FunctionNode;
				if (function != null)
				{
					if (function.Name == name)
						return function;
				}
			}

			return null;
		}

		private DiagramLink FindTrueLink(OperationNode node)
		{
			foreach (DiagramLink link in node.OutgoingLinks)
			{
				if (link.OriginAnchor == 1)
					return link;
			}

			return null;
		}

		private DiagramLink FindFalseLink(OperationNode node)
		{
			foreach (DiagramLink link in node.OutgoingLinks)
			{
				if (link.OriginAnchor == 2)
					return link;
			}

			return null;
		}

		private void ClearHighlight()
		{
			foreach (DiagramNode node in diagram.Nodes)
			{
				ScriptedNode scripted = node as ScriptedNode;
				if (scripted != null)
					scripted.HideHighlight();
			}
		}

		private void examplesListBox_DoubleClick(object sender, EventArgs e)
		{
			if (examplesListBox.SelectedIndex != -1)
			{
				string name = examplesListBox.SelectedItem.ToString();
				string content = "";
				switch (name)
				{

					case "Factorial":
						content = Properties.Resources.Factorial;
						break;

					case "Gcd":
						content = Properties.Resources.Gcd;
						break;

					case "Fibonacci":
						content = Properties.Resources.Fib;
						break;

				}

				if (!string.IsNullOrEmpty(content))
				{
					XmlDocument document = new XmlDocument();
					document.LoadXml(content);
					diagram.LoadFromXml(document);
					OnDiagramLoaded();
				}
			}
		}


		private VariableListNode variableListNode;
		private EvaluationContext evaluationContext;
		private bool tracing;
		private ScriptedNode currentNode;
	}
}