//
// Copyright (c) 2022, MindFusion LLC - Bulgaria.
//

using System;
using System.Collections.Generic;
using System.Drawing;

using MindFusion.Diagramming.Components;
using MindFusion.Drawing;


namespace MindFusion.Diagramming.WinForms.Samples.CS.Scripting
{
	public class VariableListNode : CompositeNode
	{
		private static readonly string Template = @"
			<SimplePanel>
				<Border Name=""Frame"" Background=""#80F0F0F0"" Pen=""#909090"" />
				<Border Padding=""2"">
					<GridPanel Name=""Content"">
						<GridPanel.Columns>
							<GridColumn Width=""Auto"" />
							<GridColumn Width=""Auto"" />
						</GridPanel.Columns>
						<GridPanel.Rows>
							<GridRow />
						</GridPanel.Rows>

						<Text Name=""Variable"" Text=""Variable"" TextAlignment=""Near"" />
						<Text Name=""Value"" Text=""Value"" GridColumn=""1"" TextAlignment=""Near"" />

					</GridPanel>
				</Border>
			</SimplePanel>";

		public VariableListNode(Diagram diagram)
			: base(diagram)
		{
			variables = new Dictionary<string, int>();
			font = new Font("Verdana", 5f, FontStyle.Bold, GraphicsUnit.World);

			content = XmlLoader.Load(Template, this);
			Components.Add(content);

			contentGrid = FindComponent("Content") as GridPanel;
			(FindComponent("Variable") as TextComponent).Font = font;
			(FindComponent("Value") as TextComponent).Font = font;

			Locked = true;

			AutoSize();
		}

		// do not use the built-in component serialization
		protected override bool SerializeComponents()
		{
			return false;
		}

		public void RegisterVariable(string name, object value)
		{
			TextComponent valueText = null;

			int index = -1;
			if (!variables.TryGetValue(name, out index))
			{
				index = contentGrid.Rows.Count;
				variables.Add(name, index);

				contentGrid.Rows.Add(new GridRow());

				TextComponent nameText = new TextComponent();
				nameText.Font = font;
				nameText.TextColor = Color.FromArgb(0x60, 0x60, 0x60);
				nameText.GridRow = index;
				nameText.Text = name;
				nameText.TextFormat.FormatFlags = StringFormatFlags.NoWrap;
				nameText.TextAlignment = StringAlignment.Near;
				contentGrid.Components.Add(nameText);

				valueText = new TextComponent();
				valueText.Font = font;
				valueText.TextColor = Color.Black;
				valueText.GridRow = index;
				valueText.GridColumn = 1;
				valueText.TextFormat.FormatFlags = StringFormatFlags.NoWrap;
				valueText.TextAlignment = StringAlignment.Near;
				contentGrid.Components.Add(valueText);
			}
			else
			{
				// Find the corresponding TextBox displaying the value
				foreach (TextComponent component in contentGrid.Components)
				{
					if (component.GridColumn == 1 && component.GridRow == index)
					{
						valueText = component;
						break;
					}
				}
			}

			if (valueText == null)
				return;

			valueText.Text = value != null ? value.ToString() : "(null)";

			AutoSize();
		}

		public void Clear()
		{
			variables.Clear();

			// Remove all components representing variables
			for (int i = 2; i < contentGrid.Components.Count; )
				contentGrid.Components.RemoveAt(i);

			// Remove the grid rows
			for (int i = 1; i < contentGrid.Rows.Count; )
				contentGrid.Rows.RemoveAt(i);

			AutoSize();
		}

		private void AutoSize()
		{
			IGraphics graphics = Parent.CreateMeasureGraphics();

			SizeF size = content.GetDesiredSize(new SizeF(float.PositiveInfinity, float.PositiveInfinity), graphics);

			RectangleF bounds = Bounds;
			bounds.Width = size.Width;
			bounds.Height = size.Height;
			Bounds = bounds;

			content.InvalidateLayout();

			graphics.Dispose();

			if (Resized != null)
				Resized(this, EventArgs.Empty);
		}


		public event EventHandler Resized;


		private ComponentBase content;
		private GridPanel contentGrid;
		private Dictionary<string, int> variables;
		private Font font;
	}
}
